;;; **********************************************************************
;;; Copyright (c) 89, 90, 91, 92, 93 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; send to: hkt@zkm.de
;;; **********************************************************************
;;; Cmusic syntax by Charles Baker 5/94 - based completely on hkt's code!
;;; check out new :after method for write-event (note)....
;;; and the cmus-formatting-slots hack.
;;; illustration of use in ks.lisp
;;; playcmusic shell command is for unix directory 

(in-package :stella)

(defsyntax cmusic 
  :pathname (pathname "test.sc")
  :stream-types '((cmusic-score-file "sc"))
  :nicknames '(cmus))

(defclass cmusic-score-file (event-file header-mixin)
  ((orchestra :initarg orchestra :initform NIL) ;; what orchestra uses this score file.
  (output :initarg output :initform nil) ;; what sound file to compile 
  (revtail :initarg revtail :initform 0.0)   ;; reverb decay time
  (last-time :initarg last-time :initform 0.0) ;; keeps last output time for this file.
  (syntax :initform (find-syntax ':cmusic))))

(defmethod open-event-stream ((syntax cmusic) &rest args)
  (apply #'open-event-stream (find-class 'cmusic-score-file) args))

(defmethod initialize-stream-for-processing ((stream cmusic-score-file))
  (let ((file (slot-value stream 'stream)))
    (format file "/* Stella output ~A */~%~%  " (cm::date-string)) 
    (when (slot-boundp stream 'header)
      (format file "~A~%" (slot-value stream 'header)))))

(defmethod deinitialize-stream-for-processing ((stream cmusic-score-file))
  (let ((file (slot-value stream 'stream)))
    (format file "ter ~A;~%" (if (eql (slot-value stream 'revtail) 0.0) 	
					(slot-value stream 'last-time)
 	(+ (slot-value stream 'revtail) (slot-value stream 'last-time))))))

;;; This method keeps the "last-time" slot in the cmusic-output-file uptodate, so we can
;;; correctly generate the "terminate X.X;" command to terminate at last-time + revtail
;;; revtail = global reverb decay time ..and delays, combs,etc...

(defmethod write-event :after ((object note) (stream cmusic-score-file))
(setf (slot-value stream 'last-time)(+ (slot-value object 'time) (slot-value object 'dur))))

;;; Well, this is an attempt to make it easier to define file output methods
;;; for cmusic instruments, which all have "note time name dur ...params...;"
;;; sorry for the sloppy job of consing up the slots list...how should this be done? -CB

(defmacro cmus-formatting-slots ((object stream) &rest slots)
(let* ((slots (append '(dur name time ) slots)))
(apply #'formatting-slots-internal object stream slots '(:preamble "note " :postamble ";" :eol t))))


(defmethod play-using-syntax ((syntax cmusic) file &rest args)
	(declare (ignore args))
  #+mcl (format t "Sorry, can't play cmusic from Lisp on the Mac.~%")
  #-mcl	(let* ((ofile  (slot-value (find-stream file) 'output))
		(orc  (slot-value (find-stream file) 'orchestra))
		(command (format nil "~Aunix/playcmusic  ~A ~A ~A"	
					*common-music-directory* ofile orc 
					(namestring (truename file))))
		(plcmd (format nil "~Aunix/playsnd ~A" 
					*common-music-directory* ofile)))
		(format t "Executing cmusic with:~%Output=~A ~%Instruments=~A ~%Score=~A~%" ofile orc (namestring (truename file))) 
		(cm::shell command)
		(tell-user "Playing ~A ." ofile)
		(cm::shell plcmd)))

