(in-package :stella)

;;;
;;; this files shows how to implement an object and output printing methods
;;; for csound score files.  the object defined here is for a ficticious
;;; instrument called i1. i1 inherits its rhythm and time slots from Note,
;;; so all we need to do is add duration, frequency and amp. we also add
;;; a name slot just for fun and  give it default value of "i1".
;;;

(defobject i1 (note) 
  ((name :initform "i1")
   dur
   freq
   amp))

;;;
;;; we first define a method on print-object to print useful slot information
;;; whenever the List command prints an i1 object to the terminal.  the
;;; :print-if :always directive causes any required slot that is unbound to 
;;; appear as "-unset-" when it is printed to the terminal. the :decimals 3
;;; directive means we want rounded to 3 decimals. this global setting is
;;; overridden for the Name slot, which doesnt hold numeric data.
;;; finally, we wrap the whole printout inside the printing-random-thing macro
;;;  so that our display looks like other types of data defined in the editor.
;;;  a typical i1 might therefore display as:
;;;  #<ii 10.5 1.0 440 .1 12323231>
;;;

(defmethod print-object ((object i1) stream)
  (printing-random-thing (object stream)
    ;; show rhythm and print all unbound slots as "-unset-"
    (formatting-slots (object stream :print-if :always :decimals 3)
      (name :decimals nil) rhythm dur freq amp)))

;;;
;;; this method on write-event writes our object to the open csound score 
;;; file. we get the file associated with the output stream, write the 
;;; current parameter values to it, and finish with an end-of-line. note
;;; that in this case there is no :print-if :always directive so
;;; we will get an error if we try to write an i1 object to a score file
;;; with incomplete parameter information. this is a feature not a bug
;;; since it catches the error at the earliest possible time.  note also that
;;; this method writes the time value to the score file, not the rhythm.
;;;

(defmethod write-event ((object i1) (stream csound-score-file))
  (formatting-slots (object (slot-value stream 'stream) 
                     :decimals 3 :eol t)
    (name :decimals nil) time dur freq amp))


#|
;;; after compiling and loading this file, try pasting to stella:

(thread foo ()
  (dotimes (i 20)
    (object i1 freq (between 220 440) rhythm .2 dur .1 amp .25)))
list foo
syntax csound
write foo test.sco 0


|#